% Parser.lhs
%
% This file is part of CalcCheck
%
% Copyright (c) 2011  Wolfram Kahl
% All rights reserved
%
\section{|Parser|: Processing LaTeX{} with |CalcStyle| Markup}

\begin{code}
module Parser where

import LogicE
import TeXParse
import Utils.List (mapHead)

import Text.ParserCombinators.Parsec hiding (Parser)
import Text.ParserCombinators.Parsec.Expr

import qualified Data.Char as Char
\end{code}

%{{{ \subsection{Top-Level}
\subsection{Top-Level}

\begin{code}
pFile :: Parser st [Item]
pFile = foldr ($) [] <$> many pStep
  where
    pStep = choice
      [ (:) <$> pItem
      , const id <$> docSkip
      , const id <$> line
      ]

pItem :: Parser st Item
pItem = fmap Calcs pCalcs <|> fmap Decls pDecls

docSkip  :: Parser st String
docSkip  = many1 alphaNum
          <|> texspace
          <|> pMacro0  -- harmless if tried after |pItem|
          <|> fmap (: []) (satisfy (\ c -> Char.isSeparator c || (notElem c "\\{}$" && Char.isPunctuation c)))
\end{code}
%}}}

%{{{ \subsection{Calculations}
\subsection{Calculations}

\begin{code}
pThisIs :: Parser st CalcGoal
pThisIs = macro1 "ThisIs" pGoal <?> "\\ThisIs{...}"
\end{code}

\begin{code}
pCalcStep :: Parser st CalcStep
pCalcStep = do
  macro "CalcStep"
  op <- brace (between mspace mspace (wrapped pCalcOp))
  hint <- brace' (many pTeX)
  e <- expr
  return $ CalcStep (unBin op) hint (Report [] False) e
 <?> "\\CalcStep{_}{...} <expression>"
\end{code}

\begin{code}
pCalcSteps :: ([CalcStep] -> [CalcStep]) -> Parser st ([CalcStep], CalcGoal)
pCalcSteps f = do
  s <- pCalcStep
  do  g <- pThisIs
      return (f [s], g)
   <|> pCalcSteps (f . (s :))
   <|> return (f [s], noGoal)
 <|> return (f [], noGoal)
\end{code}

\begin{code}
pCalc :: Parser st Calc
pCalc = do
  e <- expr
  sDecl <- option noGoal pThisIs
  (steps, eDecl) <- pCalcSteps id
  return $ Calc
    { calcGoal       = noGoal
    , calcGoalCheck  = []
    , calcStartDecl  = sDecl
    , calcEndDecl    = eDecl
    , calcStart      = e
    , calcSteps      = steps
    }
\end{code}

\begin{code}
pGoal :: Parser st CalcGoal
pGoal = foldr ($) noGoal <$> (spaces >> many item)
  where
    item =       ((goalAddNr <$> parenThNum) <* spaces)
            <|>  (((goalAddExpr . snd) <$> try (pTeXMathWrap expr)) <* texspace0)
            <|>  (goalAddName <$> pGoalName)
            <|>  (texspace >> return id)

pGoalName :: Parser st String
pGoalName = let
  p  =   ((:)                <$> noneOf ":($\\]}")
     <|> (showsMathWrap (:)  <$> try (pTeXMathWrap pOp))
     <|> ((++)               <$> smacro1 "textbf" pGoalName)
  f = reverse . dropWhile Char.isSpace . reverse . foldr ($) ""
  in (f <$> many1 p) <* (optional (char ':') >> spaces)
\end{code}

\begin{code}
pCalcs :: Parser st [Calc]
pCalcs = do
  (mg, cs) <- envO1 "calc" pGoal $ sepBy1 pCalc (macro' "also")
  case mg of
    Nothing -> return cs
    Just g -> case cs of
      [c] -> return [c { calcGoal = g}]
      _ -> fail "Currently, only a single calculation is allowed when proving a goal. (No ``\also''!)"
\end{code}

The |expression| environment is parsed as a trivial calculation,
for the time being.

\begin{code}
pExpression :: Parser st [Calc]
pExpression = env "expression" $ do
  e <- expr
  return . (: []) $ Calc
    { calcGoal       = noGoal
    , calcGoalCheck  = []
    , calcStartDecl  = noGoal
    , calcEndDecl    = noGoal
    , calcStart      = e
    , calcSteps      = []
    }
\end{code}
%}}}

%{{{ \subsection{Declarations}
\subsection{Declarations}

%{{{ pDecl
\begin{code}
pDecl :: Parser st Decl
pDecl = (do
    vs <- try (commasep1 var <* (macro' "declType" <|> (char ':' >> notFollowedBy (oneOf "=≡:"))))
    mspace
    t <- expr
    mr <- pRemark
    return $ DeclType vs t mr
   <?> "<var \\declType type>")
  <|> (do
    lhs <- expr
    op  <-   fmap unBin (char ':' >> choice [eq,equiv])
        <|>  (macro' "declEqu" >> return '=')
        <|>  (macro' "declequ" >> return '=')
        <|>  (macro' "declEquiv" >> return '≡')
    rhs  <- fmap Right expr
         <|> fmap (Left . show) (mbox <|> fbox)
    mr <- pRemark
    return $ DeclEqu lhs op rhs mr
   <?> "<lhs \\declEqu rhs>")
  where
    pRemark = option Nothing
      (fmap (Just . show) (macro1 "remark" (many pTeX)) <?> "\\remark{...}")
\end{code}
%}}}

%{{{ pDecls
\begin{code}
pDecls :: Parser st [Decl]
pDecls = env "decls" $ sepBy1 pDecl (macro' "also")
\end{code}
%}}}
%}}}

%{{{ \subsection{Expressions}
\subsection{Expressions}

\begin{code}
expr1 :: Parser st Expr
expr1 = buildExpressionParser ops1 expr0 <?> "<expression1>"
\end{code}

\begin{code}
expr1a :: Parser st Expr
expr1a = foldl assocChain expr1 assocOps1
\end{code}

Parsec apparently cannot handle conjunctional operators,
so we split the operator table into two halves,
and parse conjunctional chains in the middle:

\begin{code}
expr1b :: Parser st Expr
expr1b = do
  e1 <- expr1a
  ps <- many $ do
    op <- pConjOp
    e <- expr1a
    return (unBin op, e)
  return $ if null ps
    then e1
    else ConjChain e1 ps
\end{code}

\begin{code}
expr :: Parser st Expr
expr = foldl assocChain expr1b assocOps2 <?> "<expression>"
\end{code}

%{{{ assocChain
To deal with many same-precedence operators
that however allow only selected mutual associations,
we write our our processor for individual precedence levels:
\begin{code}
assocChain :: Parser st Expr  -> [(Parser st (Expr -> Expr -> Expr), Assoc)]
                              -> Parser st Expr
assocChain p0 ops = let
    p1 (p, assoc) = do
      op <- p
      return (op, (unBin op, assoc))
    pAny = foldr1 (<|>) (map p1 ops)

    assemble e0   [] = return e0
    assemble e0   [((bin, (_op, _assoc)), e1)] = return (bin e0 e1)
    assemble _e0  (((_bin, (op, AssocNone)), _e1) :
                  (((_, (op',_)) , _) : _))
       = fail $  if op == op' then op : " does not associate"
                 else op : " does not associate with " ++ [op']
    assemble e0  (((bin1, (op1, AssocLeft)), e1) :
                  ps@(((_bin2, (op2, _)) , _e2) : _))
       = if op1 == op2 || MutAssocLeft op1 op2 `elem` mutAssocs
         then assemble (bin1 e0 e1) ps
         else fail $ op1 : " does not associate with " ++ [op2]
    assemble e0  (((bin1, (op1, AssocRight)), e1) :
                  ps@(((_bin2, (op2, _)) , _e2) : _))
       = if op1 == op2
         then fmap (bin1 e0) $ assemble e1 ps
         else fail $ op1 : " does not associate with " ++ [op2]

  in do
    e0 <- p0
    ps <- many $ do
      op <- pAny
      e <- p0
      return (op, e)
    assemble e0 ps
\end{code}
%}}}

\begin{code}
pQuantVars :: Parser st [(Var, Maybe Expr)]
pQuantVars = concat <$> semisep1 pTyVars
  where
    pTyVar = do
      v <- var
      mt <- (char ':' >> mspace >> fmap Just expr) <|> return Nothing
      return (v , mt)
{-
      e <- expr
      p <- case e of
        Var v -> return (v, Nothing)
        Bin ':' (Var v) t -> return (v, Just t)
        _ -> fail "pTyVar"
      return p
-}
     <?> "<var {: type}>"
    propagate [] = []
    propagate (p : ps) = p : case snd p of
      Nothing  -> propagate ps
      Just t   -> propagate' t ps
    propagate' _ [] = []
    propagate' t ((v, Nothing) : ps) = (v, Just t) : propagate' t ps
    propagate' _ (p@(_v, Just t) : ps) = p : propagate' t ps
    pTyVars = fmap (reverse . propagate . reverse) $ commasep1 pTyVar
\end{code}

\begin{code}
pWithspot, pWith, pSpot :: Parser st ()
pWithspot = try (string "|•" >> mspace) <|> macro' "withspot" 
pWith = (charOp '|' >> mspace) <|> macro' "with" <|> macro' "mid"
pSpot = (charOp '•' >> mspace) <|> macro' "spot" <|> macro' "bullet"
\end{code}

\begin{code}
pRawQuant :: Parser st Expr
pRawQuant = try $ do
  op <- pQuantOp
  vs <- pQuantVars
  (pWithspot >> fmap (Quant op vs Nothing) expr)
   <|> do
    pWith
    mr <- option Nothing $ fmap Just expr 
    pSpot
    e <- expr
    return $ Quant op vs mr e
\end{code}


\begin{code}
pRawTuple :: Parser st Expr
pRawTuple = do
  es <- commasep1 expr
  return $ case es of
    [e] -> e
    _ -> tuple es
\end{code}

\begin{code}
pRawSetCompr :: Parser st Expr
pRawSetCompr =  do
  vs <- try pQuantVars
  (pWithspot >> fmap (SetCompr vs Nothing . Just) expr)
   <|> do
    pWith
    mr <- option Nothing $ fmap Just expr
    me <- option Nothing $ fmap Just $ pSpot >> expr
    return $ SetCompr vs mr me
\end{code}

\begin{code}
pSetParen :: Parser st Expr -> Parser st Expr
pSetParen p =  b "\\{" "\\}"
          <|>  b "\\left\\{" "\\right\\}"
          <|>  b "\\big\\{" "\\big\\}"
  where
    b l r = between (try (string l) >> mspace) (try (string r) >> mspace) p
\end{code}

\begin{code}
pSet :: Parser st Expr
pSet = pSetParen (try pRawSetCompr <|> fmap SetEnum (commasep expr))
  <?> "<set expression>"
\end{code}

\begin{code}
pInt :: Parser st String
pInt = try ((char '-' >> fmap ('-' :) digits) <|> digits <?> "<integer>")
  where digits = many1 digit <* mspace

pInteger :: Parser st Integer
pInteger = do  s <- pInt
               case reads s of
                   [(i, [])] -> return i
                   _ -> fail $ "Internal error: Integer parsing inconsistency: " ++ s
\end{code}

\begin{code}
factor :: Parser st Expr
factor  =    mparen (pRawQuant <|> pRawTuple)
        <|>  brace' expr
        <|>  do op <- pUnOpMacro; e <- brace' expr; return $ op e
        <|>  pSet
        <|>  (between (macro' "langle") (macro' "rangle") pRawTuple <?> "<pair>")
        <|>  pConst
        <|>  fmap Var var
        <|>  fmap IntLit pInteger 
\end{code}

\begin{code}
pConst :: Parser st Expr
pConst =  (macro' "false" >> return false)
     <|>  (macro' "true" >> return true)
     <|>  ((charOp '𝔹' <|> macro' "BB") >> return bool)
     <|>  ((charOp 'ℕ' <|> macro' "NN") >> return nat)
     <|>  ((charOp 'ℤ' <|> macro' "ZZ") >> return integer)
     <|>  ((charOp 'ℚ' <|> macro' "QQ") >> return rational)
     <|>  ((charOp 'ℝ' <|> macro' "RR") >> return real)
     <|>  ((charOp 'ℂ' <|> macro' "CC") >> return complex)
     <|>  ((charOp '∅' <|> macro' "emptyset") >> return emptyset)
     <|>  (macro' "fst" >> return proj1)
     <|>  (macro' "snd" >> return proj2)
     <|>  (macro' "relDom" >> return relDom)
     <|>  (macro' "relRan" >> return relRan)
     <|>  (macro' "relId" >> return relId)
     <|>  (macro' "pos" >> return positive)
     <|>  ((charOp '𝐔' <|> macro' "Universe") >> return universe)
  <?> "<constant entity>"
\end{code}

For substitution: \edcomm{WK}{|string ":="| should be discouraged.}
\begin{code}
becomes :: Parser st ()
becomes = (string ":=" >> mspace) <|> macro' "becomes"
\end{code}

\begin{code}
varName :: Parser st String
varName = do
  x <- letter
  xs <- many (alphaNum <|> char '\'')
  return $ x : xs

var :: Parser st Var
var = wrapped (do
    name <- varName <|> choice (map (\ g -> try (string $ '\\' : g)) $ filter ng greek)
    if name `elem` ["false", "true", "\\Sigma", "\\Pi"]
      then fail $ "Illegal variable name: " ++ name
      else return ()
    s <- option "" (char '_' >> fmap (\ c -> '_' : c : []) alphaNum)
    mspace
    return $ name ++ s
  <|> smacro1 "fbox" (fmap show (many pTeX))
  <|> smacro1 "mathsf" varName
  <|> smacro1 "mathrm" varName
  <|> smacro1 "mathbf" varName
  )
  <?> "<variable>"
  where
    ng x = x `notElem` ["Sigma", "Pi"]
\end{code}

\begin{code}
subst :: Parser st (Expr -> Expr)
subst = wrapped $ bracket $ do
  mspace
  vs <- commasep1 var
  becomes
  es <- commasep1 expr
  return (\ e -> Subst e vs es)
\end{code}
%}}}

%{{{ \subsection{Operators}
\subsection{Operators}

|charOp| parses single-character operators in math mode,
so space after it needs to be discarded.
|charOpNot| and |macroNot| are for operators negated using the TeX macro \verb|\not|.

\begin{code}
charOp, charOpNot :: Char -> Parser st ()
charOp c = char c >> mspace <?> [c]
charOpNot c = try (macro "not" >> charOp c)

macroNot :: String -> Parser st ()
macroNot s = try (macro "not" >> macro' s)
\end{code}

Operator parsers:
\begin{code}
eq, neq, lt, nlt, gt, ngt, leq, nleq, geq, ngeq, ele, nele, ni, nin,
   implies, nimplies, follows, nfollows, equiv, nequiv,
   subset, subseteq, supset, supseteq,
   nsubset, nsubseteq, nsupset, nsupseteq
  :: Parser st (Expr -> Expr -> Expr)
eq         =  charOp '=' >> return (≗)
lt         =  charOp '<' >> return (.<)
gt         =  charOp '>' >> return (.>)
neq        = (charOp '≠' <|> charOpNot '=' <|> macro' "neq") >> return (≠)
nlt        = (charOp '≮' <|> charOpNot '<') >> return (≮)
ngt        = (charOp '≯' <|> charOpNot '>') >> return (≯)
leq        = (charOp '≤' <|> macro' "leq") >> return (≤)
nleq       = (charOp '≰' <|> macroNot "leq") >> return (≰)
geq        = (charOp '≥' <|> macro' "geq") >> return (≥)
ngeq       = (charOp '≱' <|> macroNot "geq") >> return (≱)
ele        = (charOp '∈' <|> macro' "in") >> return (∈)
nele       = (charOp '∉' <|> macroNot "in" <|> macro' "notin") >> return (∉)
ni         = (charOp '∋' <|> macro' "ni") >> return (∋)
nin        = (charOp '∌' <|> macroNot "ni" <|> macro' "notni") >> return (∌)
implies    = (charOp '⇒' <|> macro' "implies" <|> macro' "Rightarrow") >> return (⇒)
nimplies   = (charOp '⇏' <|> macroNot "implies" <|> macro' "notimplies" <|> macroNot "Rightarrow") >> return (⇏)
follows    = (charOp '⇐' <|> macro' "follows" <|> macro' "Leftarrow") >> return (⇐)
nfollows   = (charOp '⇍' <|> macroNot "follows" <|> macro' "notfollows" <|> macroNot "Leftarrow") >> return (⇍)
equiv      = (charOp '≡' <|> macro' "equiv") >> return (≡)
nequiv     = (charOp '≢' <|> macroNot "equiv" <|> macro' "nequiv") >> return (≢)
subset     = (charOp '⊂' <|> macro' "subset") >> return (⊂)
nsubset    = (charOp '⊄' <|> macroNot "subset") >> return (⊄)
subseteq   = (charOp '⊆' <|> macro' "subseteq") >> return (⊆)
nsubseteq  = (charOp '⊈' <|> macroNot "subseteq") >> return (⊈)
supset     = (charOp '⊃' <|> macro' "supset") >> return (⊃)
nsupset    = (charOp '⊅' <|> macroNot "supset") >> return (⊅)
supseteq   = (charOp '⊇' <|> macro' "supseteq") >> return (⊇)
nsupseteq  = (charOp '⊉' <|> macroNot "supseteq") >> return (⊉)
\end{code}

|pCalcOp| contains those that can occur as first argument to \verb|\CalcStep|:

\begin{code}
pCalcOp :: Parser st (Expr -> Expr -> Expr)
pCalcOp = choice [eq, neq, lt, nlt, gt, ngt, leq, nleq, geq, ngeq,
   ele, nele, ni, nin,
   implies, nimplies, follows, nfollows, equiv, nequiv,
   subset, subseteq, supset, supseteq,
   nsubset, nsubseteq, nsupset, nsupseteq]
\end{code}

Conjunctional operators:
\begin{code}
pConjOp :: Parser st (Expr -> Expr -> Expr)
pConjOp = choice [eq, neq, lt, gt, nlt, ngt, leq, geq, nleq, ngeq, ele, nele,
   subset, subseteq, supset, supseteq, nsubset, nsubseteq, nsupset, nsupseteq]
\end{code}

Other binary operators:
\edcomm{WK}{We need a more general way to insert |wrapped| around operators.}

\begin{code}
pOtherBinOp, pBinOp,
   fcmp, circ, multiply, exponentiate, divide, add, subtr, intersect, union, land, lor,
   colon, tfun, rel, cross, star, rpseudocomp, relRRes, relLRes,
   pMax, pMin, otimes, oplus, ominus, upto
   :: Parser st (Expr -> Expr -> Expr)
pBinOp = pCalcOp <|> pOtherBinOp
pOtherBinOp = choice [fcmp, circ, multiply, exponentiate, divide, add, subtr, intersect, union, land, lor,
   colon, tfun, cross, star, rpseudocomp, relRRes, relLRes, pMax, pMin, otimes, oplus, ominus, upto]
fcmp       = (charOp '⨾' <|> macro' "fcmp" <|> macro' "comp" <|> macro' "fatsemi") >> return (⨾)
circ       = (charOp '∘' <|> macro' "circ") >> return (∘)
multiply   = wrapped ((charOp '·' <|> macro' "cdot") >> return (·))
exponentiate = (charOp '^' >> return (＾))
divide     = charOp '/' >> return (÷)
add        = wrapped (charOp '+' >> return (⊹))
subtr      = charOp '-' >> return (╌)
intersect  = (charOp '∩' <|> macro' "intersection" <|> macro' "intersect" <|> macro' "cap") >> return (∩)
union      = (charOp '∪' <|> macro' "union" <|> macro' "cup") >> return (∪)
land      = wrapped ((charOp '∧' <|> macro' "land" <|> macro' "wedge") >> return (∧))
lor       = wrapped ((charOp '∨' <|> macro' "lor" <|> macro' "vee") >> return (∨))
colon      = (charOp ':' >> return (∶))
tfun       = (charOp '→' <|> macro' "tfun" <|> macro' "rightarrow") >> return (→)
rel        = (charOp '↔' <|> macro' "rel" <|> macro' "leftrightarrow") >> return (↔)
cross      = (charOp '×' <|> macro' "cross" <|> macro' "times") >> return (×)
star       = (charOp '⋆' <|> macro' "star") >> return (⋆)
rpseudocomp = (charOp '➝' <|> macro' "rightarrowtriangle") >> return (➝)
relRRes = (charOp '＼' <|> macro' "relRRes") >> return (＼)
relLRes = (charOp '／' <|> macro' "relLRes") >> return (／)
pMin       = (charOp '↓' <|> macro' "min" <|> macro' "downarrow") >> return (↓)
pMax       = (charOp '↑' <|> macro' "max" <|> macro' "uparrow") >> return (↑)
-- |pMin       = (charOp '⊓') >> return (⊓)|
-- |pMax       = (charOp '⊔') >> return (⊔)|
otimes     = (charOp '⊗' <|> macro' "otimes") >> return (⊗)
oplus      = (charOp '⊕' <|> macro' "oplus") >> return (⊕)
ominus      = (charOp '⊖' <|> macro' "ominus") >> return (⊖)
upto        = (charOp '⋯' <|> macro' "upto") >> return (⋯)
\end{code}

For unary operators, in particular for $\lnot$,
we do want to allow nested applications without prentheses ---
apparently this cannot be done directly with Parsec's expression parser.
\begin{code}
pUnOp, pUnOps, pLnot, pSetSize, pSetCompl, pPower, pSET, pPrePlus, pPreMinus :: Parser st (Expr -> Expr)
pUnOps = foldr1 (.) <$> many1 pUnOp
pUnOp = choice [pLnot, pSetSize, pSetCompl, pPower, pSET, pPrePlus, pPreMinus]
pLnot = (charOp '¬' <|> macro' "lnot" <|> macro' "neg") >> return lnot
pSetSize = charMacro '#' >> mspace >> return setSize
pSetCompl = (charOp '∼' <|> macro' "compl") >> return setCompl
pPower = (charOp 'ℙ' <|> macro' "power" <|> macro' "pset") >> return power
pSET = macro' "SET" >> return setType
pPrePlus = charOp '+' >> return prePlus
pPreMinus = charOp '-' >> return preMinus
\end{code}

\begin{code}
pPostOp, pPostOps, pRelConvOp, pRelTCOp, pRelRTCOp :: Parser st (Expr -> Expr)
pPostOp = choice [pRelConvOp, pRelTCOp, pRelRTCOp]
pPostOps = (foldr1 (.) . reverse) <$> many1 pPostOp
pRelConvOp = (charOp '˘' <|> macro' "RELconvSymbol" <|> macro' "converse") >> return relConv
pRelTCOp = (charOp '⁺' <|> (try (string "^+" <|> string "^{+}") >> mspace)) >> return relTC
pRelRTCOp = try (string "^*" <|> string "^{*}") >> mspace >> return relRTC
\end{code}

\begin{code}
pUnOpMacro :: Parser st (Expr -> Expr)
pUnOpMacro = choice
  [ macro "RELconv" >> return relConv
  , macro "RELtc"  >> return relTC
  , macro "RELrtc" >> return relRTC
  ]
\end{code}

\begin{code}
pOp :: Parser st Op
pOp  = (unBin <$> pBinOp) <|> (unUn <$> pUnOp) <|> (unConst <$> pConst)
     <|> choice (map (snd . snd) quantOpParsers)
\end{code}

\begin{code}
pForall, pExists, pSum, pProduct, pIntersect, pUnion, pMIN, pMAX, pStar, pEquiv :: Parser st Op
pForall     = (charOp '∀' <|> macro' "forall"    ) >> return '∀'
pExists     = (charOp '∃' <|> macro' "exists"    ) >> return '∃'
pSum        = (charOp '∑' <|> macro' "Sigma"  <|> macro' "sum"    ) >> return '∑'
pProduct    = (charOp '∏' <|> macro' "Pi"     <|> macro' "prod"        ) >> return '∏'
pIntersect  = (charOp '⋂' <|> macro' "bigcap"    ) >> return '⋂'
pUnion      = (charOp '⋃' <|> macro' "bigcup"    ) >> return '⋃'
pMIN        = (charOp '⇊' <|> macro' "bigMin"    ) >> return '⇊'
pMAX        = (charOp '⇈' <|> macro' "bigMax"    ) >> return '⇈'
pStar       = (charOp '★' <|> macro' "bigstar"   ) >> return '★'
pEquiv      = (charOp '≣' <|> macro' "bigequiv"  ) >> return '≣'
\end{code}

\begin{code}
pQuantOp :: Parser st Op
pQuantOp = choice $ map h quantOpParsers
  where
    h ((_bOp, _qOp), (pB, pQ)) = wrapped (try (pQ <|> fmap unBin pB))
\end{code}

\begin{code}
quantOpParsers :: [((Op, Op), (Parser st (Expr -> Expr -> Expr), Parser st Op))]
quantOpParsers =  if map fst ps == quantOps then ps
                  else error "quantOpParsers out of sync with quantOps!"
  where
    ps =
      [ (('≡','≣'), (equiv,      pEquiv))
      , (('·','∏'), (multiply,   pProduct))
      , (('+','∑'), (add,        pSum))
      , (('∩','⋂'), (intersect,  pIntersect))
      , (('∪','⋃'), (union,      pUnion))
      , (('↓','⇊'), (pMin,       pMIN))
      , (('↑','⇈'), (pMax,       pMAX))
      , (('∧','∀'), (land,       pForall))
      , (('∨','∃'), (lor,        pExists))
      , (('⋆','★'), (star,       pStar))
      , (('⊍','⨃'), (fail "Not in input syntax: ⊍", fail "Not in input syntax: ⨃"))
      ]
\end{code}

The operator table |ops| follows \cite{Gries-Schneider-1993}
but in addition contains entries for the function arrow |tfun|,
with precedence chosen lower than $\times$,
and for ``:'' (|colon|), with second-highest precedence to
be able to handle the examples in LADM 8.1.

A problem is the treatment of non-mutually-associative operators
of the same precedence --- we want to allow:
$$\begin{array}{rcl}
  a - b + c &=& (a - b) + c
  \\
  a / b \cdot c &=& (a / b) \cdot c
\end{array}$$
but forbid
$$
   p \land q \lor r
\enskip.
$$

\edcomm{WK}{Even if we manage to get this right in parsing,
we still need to adapt also the AC-matching to take mutual associativity into account!}

For the time being, we only need associating to the left in such cases:
\begin{code}
data MutAssoc
  = MutAssocLeft Op Op
  deriving (Eq)

mutAssocs :: [MutAssoc]
mutAssocs =
  [ MutAssocLeft '+' '-'
  , MutAssocLeft '-' '+'
  , MutAssocLeft '·' '/'
  , MutAssocLeft '/' '·'
  , MutAssocLeft '≡' '≢'
  , MutAssocLeft '≢' '≡'
  ]
\end{code}

A special case is infix notation for relations:
$x\,R\,y \;\equiv\; \langle x, y \rangle \in R$,
which has dangerous overlap with non-``$.$'' function application.

\begin{code}
expr0 :: Parser st Expr
expr0 = let factor' = do
              f <- factor
              ss <- many subst
              return $ foldl (\ e s -> s e) f ss
  in do
    x <- factor'
    option x . try $ do
           r <- (factor' <?> "<infix relation>")
           y <- factor'
           return $ tuple [x,y] ∈ r
\end{code}

\begin{code}
ops1 :: OperatorTable Char st Expr
ops1 =
  -- |[Postfix subst]|
  [ [Infix colon AssocRight]
  , [Infix (charOp '.' >> return (@.)) AssocLeft
    ,Infix (mspace >> notFollowedBy (satisfy $ not . ('(' ==)) >> return (@.)) AssocLeft
    ]
  , [Postfix pPostOps]
  , [Prefix pUnOps]
  , [Infix exponentiate  AssocRight]
  ]

assocOps1 :: [[(Parser st (Expr -> Expr -> Expr), Assoc)]]
assocOps1 =
  [ [(multiply     , AssocLeft )
    ,(divide       , AssocLeft )
    ,(star         , AssocLeft )
    ]
  , [(add          , AssocLeft )
    ,(subtr        , AssocLeft )
    ,(intersect    , AssocLeft )
    ,(union        , AssocLeft )
    ,(fcmp         , AssocLeft )
    ,(circ         , AssocLeft )
    ,(cross        , AssocLeft )
    ,(rpseudocomp  , AssocRight)
    ,(relRRes      , AssocNone )
    ,(relLRes      , AssocNone )
    ,(otimes       , AssocNone )
    ,(oplus        , AssocNone )
    ,(ominus       , AssocLeft )
    ]
  , [(pMin         , AssocLeft )
    ,(pMax         , AssocLeft )
    ]
  , [(upto         , AssocNone )
    ]
  , [(tfun         , AssocRight)
    ,(rel          , AssocRight)
    ]
  ]
\end{code}

\begin{code}
assocOps2 :: [[(Parser st (Expr -> Expr -> Expr), Assoc)]]
assocOps2 =
  [ [(land      , AssocLeft )
    ,(lor       , AssocLeft )
    ]
  , [(implies   , AssocRight)
    ,(follows   , AssocLeft )
    ,(nimplies  , AssocLeft )
    ,(nfollows  , AssocLeft )
    ]
  , [(equiv     , AssocLeft )
    ,(nequiv    , AssocLeft )
    ]
  ]
\end{code}
%}}}

%{{{ \subsection{Greek Letters}
\subsection{Greek Letters}

\begin{code}
greek, greekUpper, greekLower :: [String]
greek = greekUpper ++ greekLower
greekUpper =
  [ "Alpha", "Beta", "Gamma", "Delta"
  , "Epsilon", "Zeta", "Eta", "Theta"
  , "Iota", "Kappa", "Lambda", "Mu"
  , "Nu", "Xi", "Omikron", "Pi"
  , "Rho", "Sigma", "Tau", "Ypsilon"
  , "Phi", "Chi", "Psi", "Omega"]
greekLower = "varepsilon" : "varsigma" : "varphi" : map (mapHead Char.toLower) greekUpper
\end{code}
%}}}

%{{{ EMACS lv
% Local Variables:
% folded-file: t
% fold-internal-margins: 0
% eval: (fold-set-marks "%{{{ " "%}}}")
% eval: (fold-whole-buffer)
% end:
%}}}
