% Rewrite.lhs
%
% This file is part of CalcCheck
%
% Copyright (c) 2011  Wolfram Kahl
% All rights reserved
%
\section{|Rewrite|: Rewriting of |AC| Canonical Terms}

%{{{ module Rewrite
\begin{code}
{-# LANGUAGE ScopedTypeVariables, PatternGuards #-}
module Rewrite where

import AC
import Utils.List (splitEithers, uninsert, nubOrd, nubOrdOn)

import qualified Data.Map as Subst
import qualified Data.Map as Map
import Data.Map (Map)
import qualified Data.Set as Set

import Control.Arrow (second)
import Control.Monad (liftM, MonadPlus(..))
import Control.Monad.State (evalStateT)
\end{code}
%}}}

\begin{code}
data Entry var op cnst t = Entry
  { e_tp :: TermPair var op cnst
  , e_no :: [NotOccurs var op cnst]
  , e_reason :: t
  , e_nor :: NOResult var op cnst
  }

-- type History var op cnst t = [((TermPair var op cnst, [NotOccurs var op cnst]), t)]
type History var op cnst t = [Entry var op cnst t]
type AddHistory var op cnst t = History var op cnst t -> History var op cnst t

type Result var op cnst t = [(AddHistory var op cnst t, Term var op cnst)]
type Rewrite var op cnst t = Term var op cnst -> Result var op cnst t
\end{code}

%{{{ substCheckNotOccurs
|substCheckNotOccurs| result: (internal error, (occurs, OK))
\begin{code}
type NOResult var op cnst = ([String], ([(var, Term var op cnst)], [(var, Term var op cnst)]))

emptyNOR :: NOResult var op cnst
emptyNOR = ([],([],[]))

substCheckNotOccurs ::
    (Ord var, Show var, Show op, Ord op, Show cnst, Ord cnst)
  => (Term var op cnst -> [var])
  -> Subst var op cnst -> [NotOccurs var op cnst]
  -> NOResult var op cnst

substCheckNotOccurs unVarList su nos = second splitEithers $ splitEithers (nos >>= f)
  where
    f (vs, e) = let
        e' = applySubst su e
        checkVar v = case Subst.lookup v su of
          Nothing -> [Left $ "Unmatched variable " ++ show v]
          Just q -> let
              checkVar' v' =  Right $ if varInTerm v' e' -- occurs [v'] [e']
                              then Left (v', e')
                              else Right (v', e')
            in map checkVar' $ unVarList q
      in vs >>= checkVar
\end{code}
%}}}

%{{{ rewrite
\begin{code}
rewrite :: (MonadPlus m, Ord var, Ord op, Ord cnst, Show var, Show op, Show cnst)
  => (Term var op cnst -> [var])
  -> m var
  -> (TermPair var op cnst, [NotOccurs var op cnst]) -> Term var op cnst -> m (Term var op cnst, NOResult var op cnst)
rewrite unVarList fresh rule@((lhs, rhs),nos) t@(AC op ts) =
  do  su <- match1 fresh [(lhs, t)]
      return (applySubst su rhs, substCheckNotOccurs unVarList su nos)
  `mplus`
  do  v <- liftM Var fresh
      su <- match1 fresh [(applyAC op v lhs, t)]
      return (applySubst su $ applyAC op v rhs, substCheckNotOccurs unVarList su nos)
  `mplus`
  do  (t', ts') <- foldr (mplus . return) mzero $ uninsert ts
      (u', nor) <- rewrite unVarList fresh rule t'
      return (applyAC op u' (mkAC op ts'), nor)
rewrite unVarList fresh rule@((lhs, rhs),nos) t@(C op t1 t2) =
  do  su <- match1 fresh [(lhs, t)]
      return (applySubst su rhs, substCheckNotOccurs unVarList su nos)
  `mplus`
  do  (t', mkts) <- foldr (mplus . return) mzero $ uninsert' [t1, t2]
      (u', nor) <- rewrite unVarList fresh rule t'
      return (applyCs op (mkts u'), nor)
rewrite unVarList fresh rule@((lhs, rhs),nos) t@(Appl op ts) =
  do  su <- match1 fresh [(lhs, t)]
      return (applySubst su rhs, substCheckNotOccurs unVarList su nos)
  `mplus`
  do  (t', mkts) <- foldr (mplus . return) mzero $ uninsert' ts
      (u', nor) <- rewrite unVarList fresh rule t'
      return (apply op $ mkts u', nor)
rewrite unVarList fresh ((lhs, rhs),nos) t@(Const _) =
  do  su <- match1 fresh [(lhs, t)]
      return (applySubst su rhs, substCheckNotOccurs unVarList su nos)
rewrite unVarList fresh ((lhs, rhs),nos) t@(Var _) =
  do  su <- match1 fresh [(lhs, t)]
      return (applySubst su rhs, substCheckNotOccurs unVarList su nos)
\end{code}
%}}}

%{{{ rewrite'
\begin{code}
rewrite' :: (Ord var, Ord op, Ord cnst, Show var, Show op, Show cnst)
  => (Term var op cnst -> [var])
  -> [var] -> (TermPair var op cnst, [NotOccurs var op cnst]) -> Term var op cnst -> [(Term var op cnst, NOResult var op cnst)]
rewrite' unVarList freshVars rule t = let
    fresh = mkFresh "Rewrite.rewrite"
  in nubOrd $ evalStateT (rewrite unVarList fresh rule t) freshVars
\end{code}
%}}}

%{{{ rewrite''
\begin{code}
rewrite'' :: (Ord var, Ord op, Ord cnst, Show var, Show op, Show cnst)
  => (Term var op cnst -> [var])
  -> [var] -> ((TermPair var op cnst, [NotOccurs var op cnst]), t) -> Rewrite var op cnst t
rewrite'' unVarList freshVars (rule@(tp,no), reason) t = map (\ (t', nor) -> ((e nor :), t')) $ rewrite' unVarList freshVars rule t
  where
    e nor = Entry
      { e_tp = tp
      , e_no = no
      , e_reason = reason
      , e_nor = nor
      }
\end{code}
%}}}

%{{{ level
\begin{code}
level :: forall var op cnst t . (Ord var, Ord op, Ord cnst, Show var, Show op, Show cnst)
  => (Term var op cnst ->  [Term var op cnst])
  -> [Rewrite var op cnst t]
  -> Result var op cnst t
  -> Result var op cnst t
level simp rules = foldr (\ (q, t) -> doRules q t rules) []
  where
    doRules  :: AddHistory var op cnst t -> Term var op cnst
             -> [Rewrite var op cnst t]
             -> Result var op cnst t -> Result var op cnst t
    doRules _q _t [] = id
    doRules q t (r : rs) = doRule q t r . doRules q t rs

    doRule q t r results = foldr (doRewrite q) results (simp t >>= r) -- (r t)

    doRewrite q (q', t') rr = let q'' = q . q'
      in foldr (\ u -> ((q'', u) :)) rr (simp t')
\end{code}
%}}}

%{{{ check
|simp| will be |trySubst| for now.
\begin{code}
check :: forall var op cnst reason . (Ord var, Ord op, Ord cnst, Show var, Show op, Show cnst)
  => (Term var op cnst ->  [Term var op cnst])
  -> [[Rewrite var op cnst reason]]
  -> (Term var op cnst, Term var op cnst)
  -> Maybe (History var op cnst reason)
check simp ruless (t1, t2) = let

  switch :: Result var op cnst reason -> Either (History var op cnst reason) (Result var op cnst reason)
  switch = h Set.empty
    where
      h _ [] = Right []
      h s (y@(q,t) : es) = if t2 == t then Left (q [])
          else if t `Set.member` s then h s es
               else fmap (y :) $ h (Set.insert t s) es

  -- Breadth-first search \edcomm{WK}{Add memory?}
  run [] _ps = Nothing
  run (rs : rss) ps = case switch $ level simp rs ps of
    Left prf -> Just prf
    Right ps' -> run rss ps'

  in run ruless [(id, t1)]
\end{code}
%}}}

%{{{ check2
\begin{code}
check2 :: forall var op cnst reason . (Ord var, Ord op, Ord cnst, Show var, Show op, Show cnst)
  => (Term var op cnst ->  [Term var op cnst])
  -> [Rewrite var op cnst reason]
  -> [Rewrite var op cnst reason]
  -> (Term var op cnst, Term var op cnst)
  -> Maybe (History var op cnst reason)
check2 simp tos froms (t1, t2) = let

  rewTo  :: Map (Term var op cnst) (AddHistory var op cnst reason)
         -> Map (Term var op cnst) (AddHistory var op cnst reason)
         -> [(Term var op cnst,[Rewrite var op cnst reason])]
         -> Either  ( Map (Term var op cnst) (AddHistory var op cnst reason)
                    , [(Term var op cnst,[Rewrite var op cnst reason])]
                    )
                    Ordering -- \unfinished
  rewTo s1 _s2 [] = Left (s1, [])
  rewTo s1 _s2 ((_t,[]) : tq) = Left (s1, tq)
  rewTo s1 s2 ((t,r : rs) : tq) = h s1 s2 id $ r t
    where
      h u1 _u2 c [] = Left (u1, (t,rs) : c tq)
      h u1 u2 c (p@(_q, _t') : _ps) = case rewTo1 u1 u2 p of
        Left u1' -> h u1' u2 (c . undefined) undefined -- \unfinished
        r' -> undefined r' -- \unfinished

{-
  rewTo s1 s2 r t = h s1 s2 $ r t
    where
      h s1 s2 [] = Left s1
      h s1 s2 (p : ps) = case rewTo1 s1 s2 p of
        Left s1' -> h s1' s2 ps
        r -> r
-}

  rewTo1 s1 s2 (q, t) = case Map.lookup t s2 of
    Nothing -> Left $ Map.insert t q s1
    Just q' -> Right . q $ q' []

  rewFrom  :: Map (Term var op cnst) (AddHistory var op cnst reason)
           -> Map (Term var op cnst) (AddHistory var op cnst reason)
           -> Rewrite var op cnst reason
           -> Term var op cnst
           -> Either  (Map (Term var op cnst) (AddHistory var op cnst reason))
                      (History var op cnst reason)
  rewFrom s1 s2 r t = h s1 s2 $ r t
    where
      h u1 _u2 [] = Left u1
      h u1 u2 (p : ps) = case rewFrom1 u1 u2 p of
        Left u2' -> h u1 u2' ps
        r' -> r'

  rewFrom1 s1 s2 (q, t) = case Map.lookup t s1 of
    Nothing -> Left $ Map.insert t q s2
    Just q' -> Right . q' $ q []

  switch = h Set.empty
    where
      h _ [] = Right []
      h s (y@(q,t) : es) = if t2 == t then Left (q [])
          else if t `Set.member` s then h s es
               else fmap (y :) $ h (Set.insert t s) es

  -- Breadth-first search \edcomm{WK}{Add memory?}
  run [] _ps = Nothing
  run (rs : rss) ps = case switch $ level simp rs ps of
    Left prf -> Just prf
    Right ps' -> run rss ps'

  in run (undefined tos froms rewTo rewFrom) [(id, t1)]  -- \unfinished
\end{code}
%}}}

%{{{ rewriteB
\begin{code}
rewriteB :: (MonadPlus m, Ord var, Ord op, Ord cnst, Show var, Show op, Show cnst)
  => m var -> RuleGroup var op cnst r -> Term var op cnst -> m (Term var op cnst, (TermPair var op cnst, r))
rewriteB fresh rule@(lhs, rhss) t =
  do  su <- match1 fresh [(lhs, t)]
      (rhs, r) <- foldr (mplus . return) mzero rhss
      return (applySubst su rhs, ((lhs, rhs), r))
  `mplus`
  case t of
    AC op ts ->
      do  v <- liftM Var fresh
          su <- match1 fresh [(applyAC op v lhs, t)]
          (rhs, r) <- foldr (mplus . return) mzero rhss
          return (applySubst su $ applyAC op v rhs, ((lhs, rhs), r))
      `mplus`
      do  (t', ts') <- foldr (mplus . return) mzero $ uninsert ts
          (u', r) <- rewriteB fresh rule t'
          return (applyAC op u' (mkAC op ts'), r)
    C op t1 t2 ->
      do  (t', mkts) <- foldr (mplus . return) mzero $ uninsert' [t1, t2]
          (u', r) <- rewriteB fresh rule t'
          return (applyCs op (mkts u'), r)
    Appl op ts ->
      do  (t', mkts) <- foldr (mplus . return) mzero $ uninsert' ts
          (u', r) <- rewriteB fresh rule t'
          return (apply op $ mkts u', r)
    Var _ -> mzero
    Const _ -> mzero
\end{code}
%}}}

%{{{ rewriteB'
\begin{code}
rewriteB' :: (Ord var, Ord op, Ord cnst, Show var, Show op, Show cnst)
  => [var] -> RuleGroup var op cnst r -> Term var op cnst -> [(Term var op cnst, (TermPair var op cnst, r))]
rewriteB' freshVars rule t = let
    fresh = mkFresh "Rewrite.rewriteB"
  in nubOrdOn fst $ evalStateT (rewriteB fresh rule t) freshVars
\end{code}
%}}}

%{{{ levelB
\begin{code}
levelB :: (Ord var, Ord op, Ord cnst, Show var, Show op, Show cnst)
  => [var]
  -> [RuleGroup var op cnst r]
  -> [([(TermPair var op cnst, r)] -> a, Term var op cnst)]
  -> [([(TermPair var op cnst, r)] -> a, Term var op cnst)]
levelB freshVars rules = foldr (\ (q, t) -> doRules q t rules) []
  where
    doRules _q _t [] = id
    doRules q t (rg : rgs) = doRuleGroup q t rg . doRules q t rgs

    doRuleGroup q t rg results = foldr (\ (t', r) rr -> (q . (r :), t') : rr)
      results
      (rewriteB' freshVars rg t)
\end{code}
%}}}

%{{{ checkB
\begin{code}
checkB :: (Ord var, Ord op, Ord cnst, Show var, Show op, Show cnst)
  => [var]
  -> [[(TermPair var op cnst, reason)]]
  -> (Term var op cnst, Term var op cnst)
  -> Maybe [(TermPair var op cnst, reason)]
checkB freshVars ruless (t1, t2) = let
 
  switch = h Set.empty
    where
      h _ [] = Right []
      h s (y@(q,t) : es) = if t2 == t then Left (q [])
          else if t `Set.member` s then h s es
               else fmap (y :) $ h (Set.insert t s) es

  -- Breadth-first search \edcomm{WK}{Add memory?}
  run [] _ps = Nothing
  run (rs : rss) ps = case switch $ levelB freshVars rs ps of
    Left prf -> Just prf
    Right ps' -> run rss ps'

  in run (map (groupEqus freshVars) ruless) [(id, t1)]
\end{code}
%}}}

%{{{ uninsert'
\begin{code}
uninsert' :: [a] -> [(a, (a -> [a]))]
uninsert' [] = []
uninsert' (u : us) = (u, (: us)) : h (u :) us
  where
    h _ [] = []
    h f (x : xs) = (x, f . (: xs)) : h (f . (x :)) xs
\end{code}
%}}}

%{{{ .
\begin{code}

\end{code}
%}}}

%{{{ .
\begin{code}

\end{code}
%}}}

%{{{ EMACS lv
% Local Variables:
% folded-file: t
% fold-internal-margins: 0
% eval: (fold-set-marks "%{{{ " "%}}}")
% eval: (fold-whole-buffer)
% end:
%}}}
